@@ -409,14 +409,11 @@ class Firewall(object):
         # Start transaction
         transaction = FirewallTransaction(self)
 
-        if reload:
-            self.set_policy("DROP", use_transaction=transaction)
-
         # flush rules
         self.flush(use_transaction=transaction)
 
         # If modules need to be unloaded in complete reload or if there are
-        # ipsets to get applied, limit the transaction to set_policy and flush.
+        # ipsets to get applied, limit the transaction to flush.
         #
         # Future optimization for the ipset case in reload: The transaction
         # only needs to be split here if there are conflicting ipset types in
@@ -465,9 +462,15 @@ class Firewall(object):
             log.debug1("Applying direct chains rules and passthrough rules")
             self.direct.apply_direct(transaction)
 
-            # Execute transaction
-            transaction.execute(True)
-            transaction.clear()
+            # since direct rules are easy to make syntax errors lets highlight
+            # the cause if the transaction fails.
+            try:
+                transaction.execute(True)
+                transaction.clear()
+            except FirewallError as e:
+                raise FirewallError(e.code, "Direct: %s" % (e.msg if e.msg else ""))
+            except Exception:
+                raise
 
         del transaction
 
@@ -476,11 +479,16 @@ class Firewall(object):
             tm2 = time.time()
             log.debug2("Flushing and applying took %f seconds" % (tm2 - tm1))
 
-        self._state = "RUNNING"
-
     def start(self):
-        self._start()
-        self.set_policy("ACCEPT")
+        try:
+            self._start()
+        except Exception:
+            self._state = "FAILED"
+            self.set_policy("ACCEPT")
+            raise
+        else:
+            self._state = "RUNNING"
+            self.set_policy("ACCEPT")
 
     def _loader(self, path, reader_type, combine=False):
         # combine: several zone files are getting combined into one obj
@@ -608,7 +616,7 @@ class Firewall(object):
             except FirewallError as msg:
                 log.error("Failed to load %s file '%s': %s", reader_type,
                           name, msg)
-            except Exception as msg:
+            except Exception:
                 log.error("Failed to load %s file '%s':", reader_type, name)
                 log.exception()
 
@@ -700,24 +708,24 @@ class Firewall(object):
     def get_backend_by_ipv(self, ipv):
         if self.nftables_enabled:
             return self.nftables_backend
-        if ipv == "ipv4":
+        if ipv == "ipv4" and self.ip4tables_enabled:
             return self.ip4tables_backend
-        elif ipv == "ipv6":
+        elif ipv == "ipv6" and self.ip6tables_enabled:
             return self.ip6tables_backend
-        elif ipv == "eb":
+        elif ipv == "eb" and self.ebtables_enabled:
             return self.ebtables_backend
         raise FirewallError(errors.INVALID_IPV,
-                            "'%s' is not a valid backend" % ipv)
+                            "'%s' is not a valid backend or is unavailable" % ipv)
 
     def get_direct_backend_by_ipv(self, ipv):
-        if ipv == "ipv4":
+        if ipv == "ipv4" and self.ip4tables_enabled:
             return self.ip4tables_backend
-        elif ipv == "ipv6":
+        elif ipv == "ipv6" and self.ip6tables_enabled:
             return self.ip6tables_backend
-        elif ipv == "eb":
+        elif ipv == "eb" and self.ebtables_enabled:
             return self.ebtables_backend
         raise FirewallError(errors.INVALID_IPV,
-                            "'%s' is not a valid backend" % ipv)
+                            "'%s' is not a valid backend or is unavailable" % ipv)
 
     def is_backend_enabled(self, name):
         if name == "ip4tables":
@@ -788,29 +796,29 @@ class Firewall(object):
             rules = backend.build_default_rules(self._log_denied)
             transaction.add_rules(backend, rules)
 
-        ipv6_backend = self.get_backend_by_ipv("ipv6")
-        if self.ipv6_rpfilter_enabled and \
-           "raw" in ipv6_backend.get_available_tables():
+        if self.is_ipv_enabled("ipv6"):
+            ipv6_backend = self.get_backend_by_ipv("ipv6")
+            if self.ipv6_rpfilter_enabled and \
+               "raw" in ipv6_backend.get_available_tables():
 
-            # Execute existing transaction
+                # Execute existing transaction
+                transaction.execute(True)
+                # Start new transaction
+                transaction.clear()
+
+                rules = ipv6_backend.build_rpfilter_rules(self._log_denied)
+                transaction.add_rules(ipv6_backend, rules)
+
+                # Execute ipv6_rpfilter transaction, it might fail
+                try:
+                    transaction.execute(True)
+                except FirewallError as msg:
+                    log.warning("Applying rules for ipv6_rpfilter failed: %s", msg)
+                # Start new transaction
+                transaction.clear()
+
+        if use_transaction is None:
             transaction.execute(True)
-            # Start new transaction
-            transaction.clear()
-
-            rules = ipv6_backend.build_rpfilter_rules(self._log_denied)
-            transaction.add_rules(ipv6_backend, rules)
-
-            # Execute ipv6_rpfilter transaction, it might fail
-            try:
-                transaction.execute(True)
-            except FirewallError as msg:
-                log.warning("Applying rules for ipv6_rpfilter failed: %s", msg)
-            # Start new transaction
-            transaction.clear()
-
-        else:
-            if use_transaction is None:
-                transaction.execute(True)
 
     # flush and policy
 
@@ -879,7 +887,6 @@ class Firewall(object):
                     backend.set_rule(rule, self._log_denied)
                 except Exception as msg:
                     log.debug1(traceback.format_exc())
-                    log.error("Failed to apply rules. A firewall reload might solve the issue if the firewall has been modified using ip*tables or ebtables.")
                     log.error(msg)
                     for rule in reversed(_rules[:i]):
                         try:
@@ -887,7 +894,7 @@ class Firewall(object):
                         except Exception:
                             # ignore errors here
                             pass
-                    return False
+                    raise msg
             return True
         else:
             return backend.set_rules(_rules, self._log_denied)
@@ -963,11 +970,18 @@ class Firewall(object):
         _direct_config = self.direct.get_runtime_config()
         _old_dz = self.get_default_zone()
 
+        self.set_policy("DROP")
+
         # stop
         self.cleanup()
 
-        # start
-        self._start(reload=True, complete_reload=stop)
+        start_exception = None
+        try:
+            self._start(reload=True, complete_reload=stop)
+        except Exception as e:
+            # save the exception for later, but continue restoring interfaces,
+            # etc. We'll re-raise it at the end.
+            start_exception = e
 
         # handle interfaces in the default zone and move them to the new
         # default zone if it changed
@@ -1012,6 +1026,12 @@ class Firewall(object):
         else:
             self.set_policy("ACCEPT")
 
+        if start_exception:
+            self._state = "FAILED"
+            raise start_exception
+        else:
+            self._state = "RUNNING"
+
     # STATE
 
     def get_state(self):
