@@ -323,8 +323,12 @@ class FirewallZone(object):
                     elif key == "masquerade":
                         self._masquerade(enable, _zone, zone_transaction)
                     elif key == "rules":
+                        if "mark" in obj.settings["rules"][args]:
+                            mark = obj.settings["rules"][args]["mark"]
+                        else:
+                            mark = None
                         self.__rule(enable, _zone,
-                                    Rich_Rule(rule_str=args), None,
+                                    Rich_Rule(rule_str=args), mark,
                                     zone_transaction)
                     elif key == "interfaces":
                         self._interface(enable, _zone, args, zone_transaction)
@@ -672,8 +676,7 @@ class FirewallZone(object):
         return None
 
     def __rule(self, enable, zone, rule, mark_id, zone_transaction):
-        return self._rule_prepare(enable, zone, rule, mark_id,
-                                  zone_transaction)
+        self._rule_prepare(enable, zone, rule, mark_id, zone_transaction)
 
     def add_rule(self, zone, rule, timeout=0, sender=None,
                  use_zone_transaction=None):
@@ -692,13 +695,16 @@ class FirewallZone(object):
         else:
             zone_transaction = use_zone_transaction
 
-        if _obj.applied:
-            mark = self.__rule(True, _zone, rule, None, zone_transaction)
+        if type(rule.element) == Rich_ForwardPort:
+            mark = self._fw.new_mark()
         else:
             mark = None
 
+        if _obj.applied:
+            self.__rule(True, _zone, rule, mark, zone_transaction)
+
         self.__register_rule(_obj, rule_id, mark, timeout, sender)
-        zone_transaction.add_fail(self.__unregister_rule, _obj, rule_id)
+        zone_transaction.add_fail(self.__unregister_rule, _obj, rule_id, mark)
 
         if use_zone_transaction is None:
             zone_transaction.execute(True)
@@ -720,28 +726,31 @@ class FirewallZone(object):
             raise FirewallError(errors.NOT_ENABLED,
                                 "'%s' not in '%s'" % (rule, _zone))
 
+        if "mark" in _obj.settings["rules"][rule_id]:
+            mark = _obj.settings["rules"][rule_id]["mark"]
+        else:
+            mark = None
+
         if use_zone_transaction is None:
             zone_transaction = self.new_zone_transaction(_zone)
         else:
             zone_transaction = use_zone_transaction
 
-        if "mark" in _obj.settings["rules"][rule_id]:
-            mark = _obj.settings["rules"][rule_id]["mark"]
-        else:
-            mark = None
         if _obj.applied:
             self.__rule(False, _zone, rule, mark, zone_transaction)
 
-        zone_transaction.add_post(self.__unregister_rule, _obj, rule_id)
+        zone_transaction.add_post(self.__unregister_rule, _obj, rule_id, mark)
 
         if use_zone_transaction is None:
             zone_transaction.execute(True)
 
         return _zone
 
-    def __unregister_rule(self, _obj, rule_id):
+    def __unregister_rule(self, _obj, rule_id, mark=None):
         if rule_id in _obj.settings["rules"]:
             del _obj.settings["rules"][rule_id]
+        if mark:
+            self._fw.del_mark(mark)
 
     def query_rule(self, zone, rule):
         return self.__rule_id(rule) in self.get_settings(zone)["rules"]
@@ -1554,7 +1563,7 @@ class FirewallZone(object):
         if rule.family is not None:
             ipvs = [ rule.family ]
         else:
-            ipvs = [ "ipv4", "ipv6" ]
+            ipvs = [ipv for ipv in ["ipv4", "ipv6"] if self._fw.is_ipv_enabled(ipv)]
 
         source_ipv = self._rule_source_ipv(rule.source)
         if source_ipv is not None and source_ipv != "":
@@ -1690,7 +1699,7 @@ class FirewallZone(object):
                         if backend.is_ipv_supported(ipv):
                             zone_transaction.add_post(enable_ip_forwarding, ipv)
 
-                rules = backend.build_zone_masquerade_rules(enable, zone)
+                rules = backend.build_zone_masquerade_rules(enable, zone, rule)
                 zone_transaction.add_rules(backend, rules)
 
             # FORWARD PORT
@@ -1702,18 +1711,8 @@ class FirewallZone(object):
                 for ipv in ipvs:
                     if backend.is_ipv_supported(ipv):
                         self.check_forward_port(ipv, port, protocol, toport, toaddr)
-
-                if check_single_address("ipv6", toaddr):
-                    ipv = "ipv6"
-                else:
-                    ipv = "ipv4"
-
-                if not backend.is_ipv_supported(ipv):
-                    continue
-
-                if enable:
-                    zone_transaction.add_post(enable_ip_forwarding, ipv)
-                    mark_id = self._fw.new_mark()
+                    if toaddr and enable:
+                        zone_transaction.add_post(enable_ip_forwarding, ipv)
 
                 filter_chain = "INPUT" if not toaddr else "FORWARD_IN"
 
@@ -1727,10 +1726,6 @@ class FirewallZone(object):
                                     toaddr, mark_id, rule)
                 zone_transaction.add_rules(backend, rules)
 
-                if not enable:
-                    zone_transaction.add_post(self._fw.del_mark, mark_id)
-                    mark_id = None
-
             # SOURCE PORT
             elif type(rule.element) == Rich_SourcePort:
                 port = rule.element.port
@@ -1811,6 +1806,8 @@ class FirewallZone(object):
         #
         backends_ipv = []
         for ipv in ["ipv4", "ipv6"]:
+            if not self._fw.is_ipv_enabled(ipv):
+                continue
             backend = self._fw.get_backend_by_ipv(ipv)
             if len(svc.destination) > 0:
                 if ipv in svc.destination:
@@ -1921,7 +1918,8 @@ class FirewallZone(object):
             zone_transaction.add_chain("nat", "PREROUTING")
             zone_transaction.add_chain("filter", filter_chain)
 
-        zone_transaction.add_post(enable_ip_forwarding, ipv)
+        if toaddr and enable:
+            zone_transaction.add_post(enable_ip_forwarding, ipv)
         backend = self._fw.get_backend_by_ipv(ipv)
         rules = backend.build_zone_forward_port_rules(
                             enable, zone, filter_chain, port, protocol, toport,
