@@ -290,7 +290,7 @@ class nftables(object):
             if log_denied == "off":
                 return ""
             if log_denied in ["unicast", "broadcast", "multicast"]:
-                rule[i:i+1] = ["pkttype", self._log_denied]
+                rule[i:i+1] = ["pkttype", log_denied]
             else:
                 rule.pop(i)
 
@@ -314,38 +314,20 @@ class nftables(object):
         # packets while initially starting and for panic mode. As such, using
         # hooks with a higher priority than our base chains is sufficient.
         #
-        table_chains = []
-        for table in list(IPTABLES_TO_NFT_HOOK.keys()):
-            for chain in IPTABLES_TO_NFT_HOOK[table]:
-                table_chains.append((table, chain))
-
         table_name = TABLE_NAME + "_" + "policy_drop"
 
-        def _policy_drop_helper(table, chain, family, rules):
-            _chain = "%s_%s" % (table, chain)
-            _hook = IPTABLES_TO_NFT_HOOK[table][chain][0]
-            # add hooks with priority -1, only contain drop rule
-            _priority = IPTABLES_TO_NFT_HOOK[table][chain][1] - 1
-            _add_chain = "add chain %s %s %s '{ type filter hook %s priority %d ; }'" % \
-                         (family, table_name, _chain, _hook, _priority)
-            rules.append(splitArgs(_add_chain))
-            rules.append(["add", "rule", family, table_name, _chain, "drop"])
-
         rules = []
         if policy == "DROP":
-            for family in ["inet", "ip", "ip6"]:
-                rules.append(["add", "table", family, table_name])
+            rules.append(["add", "table", "inet", table_name])
 
-            for table,chain in table_chains:
-                if table == "nat":
-                    # nat requires two families
-                    for family in ["ip", "ip6"]:
-                        _policy_drop_helper(table, chain, family, rules)
-                else:
-                    _policy_drop_helper(table, chain, "inet", rules)
+            # To drop everything we need to use the "raw" priority. These occur
+            # before conntrack, mangle, nat, etc
+            for hook in ["prerouting", "output"]:
+                _add_chain = "add chain inet %s %s_%s '{ type filter hook %s priority %d ; policy drop ; }'" % \
+                             (table_name, "raw", hook, hook, -300 + NFT_HOOK_OFFSET - 1)
+                rules.append(splitArgs(_add_chain))
         elif policy == "ACCEPT":
-            for family in ["inet", "ip", "ip6"]:
-                rules.append(["delete", "table", family, table_name])
+            rules.append(["delete", "table", "inet", table_name])
         else:
             FirewallError(UNKNOWN_ERROR, "not implemented")
 
@@ -493,6 +475,9 @@ class nftables(object):
             "OUTPUT": "oifname",
         }[chain]
 
+        if interface[len(interface)-1] == "+":
+            interface = interface[:len(interface)-1] + "*"
+
         target = DEFAULT_ZONE_TARGET.format(chain=SHORTCUTS[chain], zone=zone)
         if zone_target == DEFAULT_ZONE_TARGET:
             action = "goto"
@@ -504,10 +489,10 @@ class nftables(object):
             rule = ["add", "rule", family, "%s" % TABLE_NAME, "%s_%s_ZONES" % (table, chain)]
         else:
             rule = ["delete", "rule", family, "%s" % TABLE_NAME, "%s_%s_ZONES" % (table, chain)]
-        if interface == "+":
+        if interface == "*":
             rule += [action, "%s_%s" % (table, target)]
         else:
-            rule += [opt, interface, action, "%s_%s" % (table, target)]
+            rule += [opt, "\"" + interface + "\"", action, "%s_%s" % (table, target)]
         return [rule]
 
     def build_zone_source_address_rules(self, enable, zone, zone_target,
@@ -619,7 +604,8 @@ class nftables(object):
            target in ["ACCEPT", "REJECT", "%%REJECT%%", "DROP"] and \
            chain in ["INPUT", "FORWARD_IN", "FORWARD_OUT", "OUTPUT"]:
             rules.append(["add", "rule", family, "%s" % TABLE_NAME,
-                          "%s_%s" % (table, _zone), target.lower()])
+                          "%s_%s" % (table, _zone),
+                          target.lower() if target != "%%REJECT%%" else "%%REJECT%%"])
 
         return rules
 
@@ -681,7 +667,7 @@ class nftables(object):
         add_del = { True: "add", False: "delete" }[enable]
 
         rule = [add_del, "rule", "inet", "%s" % TABLE_NAME,
-                "%s_%s_allow" % (table, target)]
+                "%s_%s_log" % (table, target)]
         rule += rule_fragment + ["log"]
         if rich_rule.log.prefix:
             rule += ["prefix", "\"%s\"" % rich_rule.log.prefix]
@@ -698,7 +684,7 @@ class nftables(object):
         add_del = { True: "add", False: "delete" }[enable]
 
         rule = [add_del, "rule", "inet", "%s" % TABLE_NAME,
-                "%s_%s_allow" % (table, target)]
+                "%s_%s_log" % (table, target)]
         rule += rule_fragment + ["log", "level", "audit"]
         rule += self._rich_rule_limit_fragment(rich_rule.audit.limit)
 
@@ -722,10 +708,10 @@ class nftables(object):
             chain = "%s_%s_deny" % (table, target)
             rule_action = ["drop"]
         elif type(rich_rule.action) == Rich_Mark:
-            table = "mangle"
-            chain = "%s_%s_allow" % (table, target)
             target = DEFAULT_ZONE_TARGET.format(chain=SHORTCUTS["PREROUTING"],
                                                 zone=zone)
+            table = "mangle"
+            chain = "%s_%s_allow" % (table, target)
             rule_action = ["meta", "mark", "set", rich_rule.action.set]
         else:
             raise FirewallError(INVALID_RULE,
@@ -753,7 +739,7 @@ class nftables(object):
             return []
 
         rule_fragment = []
-        if check_single_address("ipv4", rich_dest.addr):
+        if check_address("ipv4", rich_dest.addr):
             rule_fragment += ["ip"]
         else:
             rule_fragment += ["ip6"]
@@ -803,7 +789,7 @@ class nftables(object):
         if rich_rule:
             rule_fragment += self._rich_rule_family_fragment(rich_rule.family)
         if destination:
-            if check_single_address("ipv4", destination):
+            if check_address("ipv4", destination):
                 rule_fragment += ["ip"]
             else:
                 rule_fragment += ["ip6"]
@@ -812,7 +798,8 @@ class nftables(object):
             rule_fragment += self._rich_rule_destination_fragment(rich_rule.destination)
             rule_fragment += self._rich_rule_source_fragment(rich_rule.source)
         rule_fragment += [proto, "dport", "%s" % portStr(port, "-")]
-        rule_fragment += ["ct", "state", "new,untracked"]
+        if not rich_rule or type(rich_rule.action) != Rich_Mark:
+            rule_fragment += ["ct", "state", "new,untracked"]
 
         rules = []
         if rich_rule:
@@ -835,7 +822,7 @@ class nftables(object):
         if rich_rule:
             rule_fragment += self._rich_rule_family_fragment(rich_rule.family)
         if destination:
-            if check_single_address("ipv4", destination):
+            if check_address("ipv4", destination):
                 rule_fragment += ["ip"]
             else:
                 rule_fragment += ["ip6"]
@@ -845,7 +832,8 @@ class nftables(object):
             rule_fragment += self._rich_rule_destination_fragment(rich_rule.destination)
             rule_fragment += self._rich_rule_source_fragment(rich_rule.source)
         rule_fragment = ["meta", "l4proto", protocol]
-        rule_fragment += ["ct", "state", "new,untracked"]
+        if not rich_rule or type(rich_rule.action) != Rich_Mark:
+            rule_fragment += ["ct", "state", "new,untracked"]
 
         rules = []
         if rich_rule:
@@ -869,7 +857,7 @@ class nftables(object):
         if rich_rule:
             rule_fragment += self._rich_rule_family_fragment(rich_rule.family)
         if destination:
-            if check_single_address("ipv4", destination):
+            if check_address("ipv4", destination):
                 rule_fragment += ["ip"]
             else:
                 rule_fragment += ["ip6"]
@@ -878,7 +866,8 @@ class nftables(object):
             rule_fragment += self._rich_rule_destination_fragment(rich_rule.destination)
             rule_fragment += self._rich_rule_source_fragment(rich_rule.source)
         rule_fragment += [proto, "sport", "%s" % portStr(port, "-")]
-        rule_fragment += ["ct", "state", "new,untracked"]
+        if not rich_rule or type(rich_rule.action) != Rich_Mark:
+            rule_fragment += ["ct", "state", "new,untracked"]
 
         rules = []
         if rich_rule:
@@ -895,20 +884,25 @@ class nftables(object):
     def build_zone_helper_ports_rules(self, enable, zone, proto, port,
                                       destination, helper_name):
         add_del = { True: "add", False: "delete" }[enable]
-        target = DEFAULT_ZONE_TARGET.format(chain=SHORTCUTS["PREROUTING"],
+        target = DEFAULT_ZONE_TARGET.format(chain=SHORTCUTS["INPUT"],
                                             zone=zone)
         rule = [add_del, "rule", "inet", "%s" % TABLE_NAME,
-                "raw_%s_allow" % (target), proto]
+                "filter_%s_allow" % (target)]
         if destination:
-            if check_single_address("ipv4", destination):
+            if check_address("ipv4", destination):
                 rule += ["ip"]
             else:
                 rule += ["ip6"]
             rule += ["daddr", destination]
-        rule += ["dport", "%s" % portStr(port, "-")]
-        rule += ["ct", "helper", helper_name]
+        rule += [proto, "dport", "%s" % portStr(port, "-")]
+        rule += ["ct", "helper", "set", "\"helper-%s-%s\"" % (helper_name, proto)]
 
-        return [rule]
+        helper_object = ["ct", "helper", "inet", TABLE_NAME,
+                         "helper-%s-%s" % (helper_name, proto),
+                         "{", "type", "\"%s\"" % (helper_name), "protocol",
+                         proto, ";", "}"]
+
+        return [helper_object, rule]
 
     def _build_zone_masquerade_nat_rules(self, enable, zone, family, rich_rule=None):
         add_del = { True: "add", False: "delete" }[enable]
@@ -917,7 +911,6 @@ class nftables(object):
 
         rule_fragment = []
         if rich_rule:
-            rule_fragment += self._rich_rule_family_fragment(rich_rule.family)
             rule_fragment += self._rich_rule_destination_fragment(rich_rule.destination)
             rule_fragment += self._rich_rule_source_fragment(rich_rule.source)
 
@@ -929,10 +922,10 @@ class nftables(object):
         # nat tables needs to use ip/ip6 family
         rules = []
         if rich_rule and (rich_rule.family and rich_rule.family == "ipv6"
-           or rich_rule.source and check_address("ipv6", rich_rule.source)):
+           or rich_rule.source and check_address("ipv6", rich_rule.source.addr)):
             rules.extend(self._build_zone_masquerade_nat_rules(enable, zone, "ip6", rich_rule))
-        if rich_rule and (rich_rule.family and rich_rule.family == "ipv4"
-           or rich_rule.source and check_address("ipv4", rich_rule.source)):
+        elif rich_rule and (rich_rule.family and rich_rule.family == "ipv4"
+           or rich_rule.source and check_address("ipv4", rich_rule.source.addr)):
             rules.extend(self._build_zone_masquerade_nat_rules(enable, zone, "ip", rich_rule))
         else:
             rules.extend(self._build_zone_masquerade_nat_rules(enable, zone, "ip6", rich_rule))
@@ -998,7 +991,7 @@ class nftables(object):
            or toaddr and check_single_address("ipv6", toaddr)):
             rules.extend(self._build_zone_forward_port_nat_rules(enable, zone,
                                 protocol, mark_fragment, toaddr, toport, "ip6"))
-        if rich_rule and (rich_rule.family and rich_rule.family == "ipv4"
+        elif rich_rule and (rich_rule.family and rich_rule.family == "ipv4"
            or toaddr and check_single_address("ipv4", toaddr)):
             rules.extend(self._build_zone_forward_port_nat_rules(enable, zone,
                                 protocol, mark_fragment, toaddr, toport, "ip"))
