@@ -21,6 +21,7 @@
 
 __all__ = [ "Rich_Source", "Rich_Destination", "Rich_Service", "Rich_Port",
             "Rich_Protocol", "Rich_Masquerade", "Rich_IcmpBlock",
+            "Rich_IcmpType",
             "Rich_SourcePort", "Rich_ForwardPort", "Rich_Log", "Rich_Audit",
             "Rich_Accept", "Rich_Reject", "Rich_Drop", "Rich_Mark",
             "Rich_Limit", "Rich_Rule" ]
@@ -45,15 +46,21 @@ class Rich_Source(object):
         if self.ipset == "":
             self.ipset = None
         self.invert = invert
+        if self.addr is None and self.mac is None and self.ipset is None:
+            raise FirewallError(errors.INVALID_RULE,
+                                "no address, mac and ipset")
 
     def __str__(self):
-        if self.addr:
-            x = ' address="%s"' % self.addr
-        elif self.mac:
-            x = ' mac="%s"' % self.mac
-        elif self.ipset:
-            x = ' ipset="%s"' % self.ipset
-        return 'source%s%s' % (" NOT" if self.invert else "", x)
+        ret = 'source%s ' % (" NOT" if self.invert else "")
+        if self.addr is not None:
+            return ret + 'address="%s"' % self.addr
+        elif self.mac is not None:
+            return ret + 'mac="%s"' % self.mac
+        elif self.ipset is not None:
+            return ret + 'ipset="%s"' % self.ipset
+        else:
+            raise FirewallError(errors.INVALID_RULE,
+                                "no address, mac and ipset")
 
 class Rich_Destination(object):
     def __init__(self, addr, invert=False):
@@ -105,6 +112,13 @@ class Rich_IcmpBlock(object):
     def __str__(self):
         return 'icmp-block name="%s"' % (self.name)
 
+class Rich_IcmpType(object):
+    def __init__(self, name):
+        self.name = name
+
+    def __str__(self):
+        return 'icmp-type name="%s"' % (self.name)
+
 class Rich_ForwardPort(object):
     def __init__(self, port, protocol, to_port, to_address):
         self.port = port
@@ -319,14 +333,15 @@ class Rich_Rule(object):
                     raise FirewallError(errors.INVALID_RULE, "bad attribute '%s'" % attr_name)
             else:             # element
                 if element in ['rule', 'source', 'destination', 'protocol',
-                               'service', 'port', 'icmp-block', 'masquerade',
+                               'service', 'port', 'icmp-block', 'icmp-type', 'masquerade',
                                'forward-port', 'source-port', 'log', 'audit',
                                'accept', 'drop', 'reject', 'mark', 'limit', 'not', 'NOT', 'EOL']:
                     if element == 'source' and self.source:
                         raise FirewallError(errors.INVALID_RULE, "more than one 'source' element")
                     elif element == 'destination' and self.destination:
                         raise FirewallError(errors.INVALID_RULE, "more than one 'destination' element")
-                    elif element in ['protocol', 'service', 'port', 'icmp-block',
+                    elif element in ['protocol', 'service', 'port',
+                                     'icmp-block', 'icmp-type',
                                      'masquerade', 'forward-port',
                                      'source-port'] and self.element:
                         raise FirewallError(errors.INVALID_RULE, "more than one element. There cannot be both '%s' and '%s' in one rule." % (element, self.element))
@@ -370,7 +385,7 @@ class Rich_Rule(object):
                 elif element in ['not', 'NOT']:
                     attrs['invert'] = True
                 else:
-                    self.source = Rich_Source(attrs.get('address'), attrs.get('mac'), attrs.get('ipset'), attrs.get('invert'))
+                    self.source = Rich_Source(attrs.get('address'), attrs.get('mac'), attrs.get('ipset'), attrs.get('invert', False))
                     in_elements.pop() # source
                     attrs.clear()
                     index = index -1 # return token to input
@@ -410,6 +425,12 @@ class Rich_Rule(object):
                     in_elements.pop() # icmp-block
                 else:
                     raise FirewallError(errors.INVALID_RULE, "invalid 'icmp-block' element")
+            elif in_element == 'icmp-type':
+                if attr_name == 'name':
+                    self.element = Rich_IcmpType(attr_value)
+                    in_elements.pop() # icmp-type
+                else:
+                    raise FirewallError(errors.INVALID_RULE, "invalid 'icmp-type' element")
             elif in_element == 'masquerade':
                 self.element = Rich_Masquerade()
                 in_elements.pop()
@@ -527,10 +548,14 @@ class Rich_Rule(object):
                     raise FirewallError(errors.INVALID_FAMILY)
                 if self.source.mac is not None:
                     raise FirewallError(errors.INVALID_RULE, "address and mac")
+                if self.source.ipset is not None:
+                    raise FirewallError(errors.INVALID_RULE, "address and ipset")
                 if not functions.check_address(self.family, self.source.addr):
                     raise FirewallError(errors.INVALID_ADDR, str(self.source.addr))
 
             elif self.source.mac is not None:
+                if self.source.ipset is not None:
+                    raise FirewallError(errors.INVALID_RULE, "mac and ipset")
                 if not functions.check_mac(self.source.mac):
                     raise FirewallError(errors.INVALID_MAC, str(self.source.mac))
 
@@ -561,7 +586,7 @@ class Rich_Rule(object):
         elif type(self.element) == Rich_Port:
             if not functions.check_port(self.element.port):
                 raise FirewallError(errors.INVALID_PORT, self.element.port)
-            if self.element.protocol not in [ "tcp", "udp" ]:
+            if self.element.protocol not in [ "tcp", "udp", "sctp", "dccp" ]:
                 raise FirewallError(errors.INVALID_PROTOCOL, self.element.protocol)
 
         # protocol
@@ -585,11 +610,18 @@ class Rich_Rule(object):
             if self.action:
                 raise FirewallError(errors.INVALID_RULE, "icmp-block and action")
 
+        # icmp-type
+        elif type(self.element) == Rich_IcmpType:
+            # icmp type availability needs to be checked in Firewall, here is no
+            # knowledge about this, therefore only simple check
+            if self.element.name is None or len(self.element.name) < 1:
+                raise FirewallError(errors.INVALID_ICMPTYPE, str(self.element.name))
+
         # forward-port
         elif type(self.element) == Rich_ForwardPort:
             if not functions.check_port(self.element.port):
                 raise FirewallError(errors.INVALID_PORT, self.element.port)
-            if self.element.protocol not in [ "tcp", "udp" ]:
+            if self.element.protocol not in [ "tcp", "udp", "sctp", "dccp" ]:
                 raise FirewallError(errors.INVALID_PROTOCOL, self.element.protocol)
             if self.element.to_port == "" and self.element.to_address == "":
                 raise FirewallError(errors.INVALID_PORT, self.element.to_port)
@@ -609,7 +641,7 @@ class Rich_Rule(object):
         elif type(self.element) == Rich_SourcePort:
             if not functions.check_port(self.element.port):
                 raise FirewallError(errors.INVALID_PORT, self.element.port)
-            if self.element.protocol not in [ "tcp", "udp" ]:
+            if self.element.protocol not in [ "tcp", "udp", "sctp", "dccp" ]:
                 raise FirewallError(errors.INVALID_PROTOCOL, self.element.protocol)
 
         # other element and not empty?
